/*
** spost.c -- feed messages to sendmail
**
** This is a simpler, faster, replacement for "post" for use
** when "sendmail" is the transport system.
**
** This code is Copyright (c) 2002, by the authors of nmh.  See the
** COPYRIGHT file in the root directory of the nmh distribution for
** complete copyright information.
*/

#include <h/mh.h>
#include <signal.h>
#include <h/addrsbr.h>
#include <h/aliasbr.h>
#include <h/dropsbr.h>
#include <h/tws.h>
#include <h/utils.h>

#define MAX_SM_FIELD 1476  /* < largest hdr field sendmail will accept */

static struct swit switches[] = {
#define VERBSW  0
	{ "verbose", 0 },
#define NVERBSW  1
	{ "noverbose", 2 },
#define ALIASW  2
	{ "alias aliasfile", 0 },
#define NALIASW  3
	{ "noalias", 2 },
#define VERSIONSW  4
	{ "Version", 0 },
#define HELPSW  5
	{ "help", 0 },
#define DEBUGSW  6
	{ "debug", -5 },
#define DISTSW  7
	{ "dist", -4 },  /* interface from dist */
#define LIBSW  8
	{ "library directory", -7 },
	{ NULL, 0 }
};


/* flags for headers->flags */
#define HNOP  0x0000  /* just used to keep .set around */
#define HBAD  0x0001  /* bad header - don't let it through */
#define HADR  0x0002  /* header has an address field */
#define HSUB  0x0004  /* Subject: header */
#define HTRY  0x0008  /* try to send to addrs on header */
#define HBCC  0x0010  /* don't output this header */
#define HFCC  0x0020  /* FCC: type header */
#define HIGN  0x0040  /* ignore this header */

/* flags for headers->set */
#define MFRM  0x0001  /* we've seen a From: */
#define MDAT  0x0002  /* we've seen a Date: */
#define MRFM  0x0004  /* we've seen a Resent-From: */
#define MVIS  0x0008  /* we've seen sighted addrs */
#define MINV  0x0010  /* we've seen blind addrs */
#define MRDT  0x0020  /* we've seen a Resent-Date: */

struct headers {
	char *value;
	unsigned int flags;
	unsigned int set;
};

static struct headers NHeaders[] = {
	{ "Return-Path", HBAD, 0 },
	{ "Received", HBAD, 0 },
	{ "Reply-To", HADR, 0 },
	{ "From", HADR, MFRM },
	{ "Sender", HADR|HBAD, 0 },
	{ "Date", HNOP, MDAT },
	{ "Subject", HSUB, 0 },
	{ "To", HADR|HTRY, MVIS },
	{ "Cc", HADR|HTRY, MVIS },
	{ "Bcc", HADR|HTRY|HBCC, MINV },
	{ "Message-Id", HBAD, 0 },
	{ "Fcc", HFCC, 0 },
	{ "Envelope-From", HIGN, 0 },
	{ NULL, 0, 0 }
};

static struct headers RHeaders[] = {
	{ "Resent-Reply-To",   HADR, 0 },
	{ "Resent-From", HADR, MRFM },
	{ "Resent-Sender", HADR|HBAD, 0 },
	{ "Resent-Date", HNOP, MRDT },
	{ "Resent-Subject", HSUB, 0 },
	{ "Resent-To", HADR|HTRY, MVIS },
	{ "Resent-Cc", HADR|HTRY, MVIS },
	{ "Resent-Bcc", HADR|HTRY|HBCC, MINV },
	{ "Resent-Message-Id", HBAD, 0 },
	{ "Resent-Fcc", HFCC, 0 },
	{ "Reply-To", HADR, 0 },
	{ "Fcc", HIGN, 0 },
	{ "Envelope-From", HIGN, 0 },
	{ NULL, 0, 0 }
};


static int badmsg = 0;
static int verbose = 0;
static int debug = 0;
static int aliasflg = 0;  /* if going to process aliases */

static unsigned msgflags = 0;  /* what we've seen */

static enum {
	normal, resent
} msgstate = normal;

static char *tmpfil;

static char *subject = NULL;  /* the subject field for BCC'ing */
static char fccs[BUFSIZ] = "";
struct mailname *bccs = NULL;  /* list of the bcc recipients */

static struct headers *hdrtab;  /* table for the message we're doing */
static FILE *out;  /* output (temp) file */

/*
** static prototypes
*/
static void putfmt(char *, char *, FILE *);
static void finish_headers(FILE *);
static int get_header(char *, struct headers *);
static void putadr(char *, struct mailname *);
static int putone(char *, int, int);
static void process_fcc(char *);
static void fcc(char *, char *);
static void process_bccs(char *);


int
main(int argc, char **argv)
{
	int state, compnum;
	char *cp, *msg = NULL, **argp, **arguments;
	char *sargv[16], buf[BUFSIZ], name[NAMESZ];
	FILE *in;

	setlocale(LC_ALL, "");
	invo_name = mhbasename(argv[0]);

	/* foil search of user profile/context */
	if (context_foil(NULL) == -1)
		done(1);

	arguments = getarguments(invo_name, argc, argv, 0);
	argp = arguments;

	while ((cp = *argp++)) {
		if (*cp == '-') {
			switch (smatch(++cp, switches)) {
			case AMBIGSW:
				ambigsw(cp, switches);
				done(1);
			case UNKWNSW:
				adios(NULL, "-%s unknown", cp);

			case HELPSW:
				snprintf(buf, sizeof(buf),
						"%s [switches] file",
						invo_name);
				print_help(buf, switches, 1);
				done(1);
			case VERSIONSW:
				print_version(invo_name);
				done(1);

			case DEBUGSW:
				debug++;
				continue;

			case DISTSW:
				msgstate = resent;
				continue;

			case VERBSW:
				verbose++;
				continue;
			case NVERBSW:
				verbose = 0;
				continue;

			case ALIASW:
				if (!(cp = *argp++) || *cp == '-')
					adios(NULL, "missing argument to %s",
							argp[-2]);
				aliasflg = 1;
				if ((state = alias(cp)) != AK_OK)
					adios(NULL, "aliasing error in file %s - %s", cp, akerror(state));
				continue;
			case NALIASW:
				aliasflg = 0;
				continue;

			case LIBSW:
				if (!(cp = *argp++) || *cp == '-')
					adios(NULL, "missing argument to %s",
							argp[-2]);
				/* create a minimal context */
				if (context_foil(cp) == -1)
					done(1);
				continue;
			}
		}
		if (msg)
			adios(NULL, "only one message at a time!");
		else
			msg = cp;
	}

	if (!msg)
		adios(NULL, "usage: %s [switches] file", invo_name);

	if ((in = fopen(msg, "r")) == NULL)
		adios(msg, "unable to open");

	if (debug) {
		verbose++;
		out = stdout;
	} else {
		tmpfil = getcpy(m_mktemp2("/tmp/", invo_name, NULL, &out));
	}

	hdrtab = (msgstate == normal) ? NHeaders : RHeaders;

	for (compnum = 1, state = FLD;;) {
		switch (state = m_getfld(state, name, buf, sizeof(buf), in)) {
		case FLD:
			compnum++;
			putfmt(name, buf, out);
			continue;

		case FLDPLUS:
			compnum++;
			cp = add(buf, cp);
			while (state == FLDPLUS) {
				state = m_getfld(state, name, buf,
						sizeof(buf), in);
				cp = add(buf, cp);
			}
			putfmt(name, cp, out);
			free(cp);
			continue;

		case BODY:
			finish_headers(out);
			fprintf(out, "\n%s", buf);
			while (state == BODY) {
				state = m_getfld(state, name, buf,
						sizeof(buf), in);
				fputs(buf, out);
			}
			break;

		case FILEEOF:
			finish_headers(out);
			break;

		case LENERR:
		case FMTERR:
			adios(NULL, "message format error in component #%d",
					compnum);

		default:
			adios(NULL, "getfld() returned %d", state);
		}
		break;
	}
	fclose(in);

	if (debug) {
		/* stop here */
		done(0);
	}

	fclose(out);

	/* process Fcc */
	if (*fccs) {
		fcc(tmpfil, fccs);
	}

	if (bccs) {
		process_bccs(tmpfil);
		if (!(msgflags & MVIS)) {
			/* only Bcc rcpts: we're finished here */
			unlink(tmpfil);
			exit(0);
		}
	}

	/*
	** re-open the temp file, unlink it and exec sendmail, giving it
	** the msg temp file as std in.
	*/
	if (!freopen(tmpfil, "r", stdin)) {
		adios(tmpfil, "can't reopen for sendmail");
	}
	unlink(tmpfil);

	argp = sargv;
	*argp++ = "send-mail";
	*argp++ = "-m";  /* send to me too */
	*argp++ = "-t";  /* read msg for recipients */
	*argp++ = "-i";  /* don't stop on "." */
	if (verbose) {
		*argp++ = "-v";
	}
	*argp = NULL;
	execv(sendmail, sargv);
	adios(sendmail, "can't exec");
	return -1;
}


/* DRAFT GENERATION */

static void
putfmt(char *name, char *str, FILE *out)
{
	int i;
	char *cp;
	struct headers *hdr;

	/* remove all leading whitespace (even newlines) */
	while (*str==' ' || *str=='\t' || *str=='\n') {
		str++;
	}

	if ((i = get_header(name, hdrtab)) == NOTOK) {
		/* no header we would care for: push it through */
		fprintf(out, "%s: %s", name, str);
		return;
	}
	/* it's one of the interesting headers */
	hdr = &hdrtab[i];

	if (hdr->flags & HIGN || !*str) {
		return;
	}

	if (hdr->flags & HBAD) {
		advise(NULL, "illegal header line -- %s:", name);
		badmsg++;
		return;
	}

	msgflags |= hdr->set;

	if (hdr->flags & HFCC) {
		process_fcc(str);
		return;
	}

	if (hdr->flags & HBCC) {
		struct mailname *mp = NULL;

		/* Create list of Bcc addrs. */
		while ((cp = getname(str))) {
			mp = getm(cp, NULL, 0, AD_HOST, NULL);
			mp->m_next = bccs;  /* push */
			bccs = mp;
		}
		return;
	}

	if (aliasflg && hdr->flags & HTRY) {
		/*
		** This header contains address(es) that we have to do
		** alias expansion on.  Because of the saved state in
		** getname we have to put all the addresses into a list.
		**/
		struct mailname *f = NULL;
		struct mailname *mp = NULL;

		while ((cp = getname(str))) {
			mp = getm(cp, NULL, 0, AD_HOST, NULL);
			if (!f) {
				f = mp;
				mp->m_next = mp;
			} else {
				mp->m_next = f->m_next;
				f->m_next = mp;
				f = mp;
			}
		}
		f = mp->m_next;
		mp->m_next = NULL;
		/* Now munch on the list, possibly expanding aliases */
		putadr(name, f);
		return;
	}

	/*
	** The author(s) of spost decided that alias substitution wasn't
	** necessary for the non-HTRY headers.	Unfortunately, one of
	** those headers is "From:", and having alias substitution work on
	** that is extremely useful for someone with a lot of POP3 email
	** accounts or aliases.  post supports aliasing of "From:"...
	**
	** Since "From:"-processing is incompletely implemented in this
	** unsupported and undocumented spost backend, I'm not going
	** to take the time to implement my new draft-From:-based email
	** address masquerading.  If I do ever implement it here, I'd almost
	** certainly want to implement "From:" line alias processing as well.
	** -- Dan Harkless <dan-nmh@dilvish.speed.net>
	*/
	/*
	** Although there is no masquerading anymore in mmh, we might want
	** to have aliasing of From: addresses. Think about it.
	** -- meillo@marmaro.de  2012-02
	*/

	if (hdr->flags & HSUB) {
		subject = getcpy(str);
	}
	fprintf(out, "%s: %s", name, str);
}


/*
** Add yet missing headers.
*/
static void
finish_headers(FILE *out)
{
	char *cp;
	char from[BUFSIZ];  /* my network address */
	char signature[BUFSIZ];  /* my signature */
	char *resentstr = (msgstate == resent) ? "Resent-" : "";

	if (!(msgflags & MDAT)) {
		fprintf(out, "%sDate: %s\n", resentstr, dtimenow());
	}

	strncpy(from, getusername(), sizeof(from));
	if ((cp = getfullname()) && *cp) {
		snprintf(signature, sizeof(signature), "%s <%s>", cp, from);
	} else {
		snprintf(signature, sizeof(signature), "%s", from);
	}
	if (!(msgflags & MFRM)) {
		fprintf(out, "%sFrom: %s\n", resentstr, signature);
	} else {
		/* In case the From: header contains multiple addresses. */
		fprintf(out, "%sSender: %s\n", resentstr, from);
	}
	if (!(msgflags & MVIS)) {
		fprintf(out, "%sBcc: Blind Distribution List: ;\n", resentstr);
	}
	if (badmsg) {
		unlink(tmpfil);
		adios(NULL, "re-format message and try again");
	}
}


/*
** Return index of the requested header in the table, or NOTOK if missing.
*/
static int
get_header(char *header, struct headers *table)
{
	struct headers *h;

	for (h=table; h->value; h++) {
		if (mh_strcasecmp(header, h->value)==0) {
			return (h - table);
		}
	}

	return NOTOK;
}


/*
** output the address list for header "name".  The address list
** is a linked list of mailname structs.  "nl" points to the head
** of the list.  Alias substitution should be done on nl.
*/
static void
putadr(char *name, struct mailname *nl)
{
	struct mailname *mp, *mp2;
	int linepos;
	char *cp;
	int namelen;

	fprintf(out, "%s: ", name);
	namelen = strlen(name) + 2;
	linepos = namelen;

	for (mp = nl; mp; ) {
		if (linepos > MAX_SM_FIELD) {
			fprintf(out, "\n%s: ", name);
			linepos = namelen;
		}
		if (mp->m_nohost) {
			/* a local name - see if it's an alias */
			cp = akvalue(mp->m_mbox);
			if (cp == mp->m_mbox) {
				/* wasn't an alias - use what the user typed */
				linepos = putone(mp->m_text, linepos, namelen);
			} else {
				/* an alias - expand it */
				while ((cp = getname(cp))) {
					if (linepos > MAX_SM_FIELD) {
						fprintf(out, "\n%s: ", name);
						linepos = namelen;
					}
					mp2 = getm(cp, NULL, 0, AD_HOST, NULL);
					if (akvisible()) {
						mp2->m_pers = getcpy(mp->m_mbox);
						linepos = putone(adrformat(mp2), linepos, namelen);
					} else {
						linepos = putone(mp2->m_text,
								linepos,
								namelen);
					}
					mnfree(mp2);
				}
			}
		} else {
			/* not a local name - use what the user typed */
			linepos = putone(mp->m_text, linepos, namelen);
		}
		mp2 = mp;
		mp = mp->m_next;
		mnfree(mp2);
	}
	putc('\n', out);
}

static int
putone(char *adr, int pos, int indent)
{
	register int len;
	static int linepos;

	len = strlen(adr);
	if (pos == indent)
		linepos = pos;
	else if (linepos+len > OUTPUTLINELEN) {
		fprintf(out, ",\n%*s", indent, "");
		linepos = indent;
		pos += indent + 2;
	} else {
		fputs(", ", out);
		linepos += 2;
		pos += 2;
	}
	fputs(adr, out);

	linepos += len;
	return (pos+len);
}


static void
process_fcc(char *str)
{
	char *cp, *pp;
	int state = 0;

	if (strlen(str)+strlen(fccs) > sizeof fccs /2) {
		adios(NULL, "Too much Fcc data");
	}
	/* todo: better have three states: SEPARATOR, PLUS, WORD */
	for (cp=pp=str; *cp; cp++) {
		switch (*cp) {
		case ' ':
		case '\t':
		case '\n':
		case ',':
			if (state != 0) {
				state = 0;
				*cp = '\0';
				if (*pp=='+' || *pp=='@') {
					strcat(fccs, " ");
				} else {
					strcat(fccs, " +");
				}
				strcat(fccs, pp);
			}
			break;
		default:
			if (state == 0) {
				state = 1;
				pp = cp;
			}
			break;
		}
	}
}


static void
fcc(char *file, char *folders)
{
	int status;
	char cmd[BUFSIZ];

	if (verbose) {
		printf("%sFcc: %s\n", msgstate == resent ? "Resent-" : "",
				folders);
		fflush(stdout);
	}
	if (100+strlen(file)+strlen(folders) > sizeof cmd) {
		adios(NULL, "Too much Fcc data");
	}
	/* hack: read from /dev/null and refile(1) won't question us */
	snprintf(cmd, sizeof cmd, "</dev/null refile -link -file '%s' %s",
			file, folders);
	status = system(cmd);
	if (status == -1) {
		fprintf(stderr, "Skipped %sFcc %s: unable to system().\n",
				msgstate == resent ? "Resent-" : "", folders);
	} else if (status != 0) {
		fprintf(stderr, "%sFcc %s: Problems occured.\n",
				msgstate == resent ? "Resent-" : "", folders);
	}
}


/* BCC GENERATION */

static void
process_bccs(char *origmsg)
{
	char *bccdraft = NULL;
	char buf[BUFSIZ];
	struct mailname *mp = NULL;
	FILE *out = NULL;

	for (mp=bccs; mp; mp=mp->m_next) {
		/*
		** Note: This draft file will be left existing by send(1),
		** although renamed with backup prefix.
		** TODO: We should have it removed eventually.
		*/
		bccdraft = getcpy(m_mktemp2("/tmp/", invo_name, NULL, &out));
		fprintf(out, "To: %s\n", mp->m_text);
		fprintf(out, "Subject: [BCC] %s", subject ? subject : "");
		fprintf(out, "%s: %s\n", attach_hdr, origmsg);
		fprintf(out, "------------\n");
		fclose(out);

		snprintf(buf, sizeof buf, "send %s", bccdraft);
		if (system(buf) != 0) {
			admonish(invo_name, "Problems to send Bcc to %s",
					mp->m_text);
			unlink(bccdraft);
		}
		/* TODO: unlink renamed bcc draft after send(1) */
	}
}
